#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import gatetools
import gatetools.phsp as phsp
import click
import itk
import numpy as np
import os
import scipy


def compute_positions(phsp, keys, e_min):
    # get speed of light in mm/s
    c = scipy.constants.speed_of_light * 1000  # in mm

    # loop on x ; check energy
    positions = phsp[:, keys.index("PostPosition_X") : keys.index("PostPosition_Z") + 1]
    directions = phsp[
        :, keys.index("PostDirection_X") : keys.index("PostDirection_Z") + 1
    ]
    # time is nanosecond so 1e9 to get in sec
    times = phsp[:, keys.index("TimeFromBeginOfEvent")] / 1e9
    energies = phsp[:, keys.index("KineticEnergy")]

    # filter according to E ?
    mask = energies > e_min
    positions = positions[mask]
    directions = directions[mask]
    times = times[mask]

    # normalise direction (not needed)
    # print('d', directions)
    # directions = directions / np.linalg.norm(directions, axis=1)
    # print('dn', directions)

    # output
    emissions = np.zeros_like(positions)

    # loop on event
    for pos, dir, t, E, p in zip(positions, directions, times, energies, emissions):
        # distance traveled according to time
        l = t * c
        # consider vector from position to dir
        # (+= is needed to keep initial pointer)
        p += pos + l * -dir

    return emissions


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("file_inputs", nargs=-1)
@click.option("--output", "-o", default="AUTO", help="output image (mhd)")
@click.option(
    "--output_folder", "-f", default=".", help="output folder, auto image name (mhd)"
)
@click.option("-n", default="-1", help="Number of samples")
@click.option(
    "--shuffle",
    "-s",
    is_flag=True,
    default=False,
    help="Shuffle the n samples (slow if file is large)",
)
def go(file_inputs, n, output, output_folder, shuffle):
    for file_input in file_inputs:
        go_one(file_input, n, output, output_folder, shuffle)


def go_one(file_input, n, output, output_folder, shuffle):
    n = int(float(n))

    # read phsp
    phsp, keys, m = gatetools.phsp.load(file_input, nmax=n, shuffle=shuffle)
    if n == -1:
        n = m

    print(f"Load {file_input} {n}/{m}")
    print(f"Keys {keys}")

    # retrieve 3D position
    positions = compute_positions(phsp, keys, e_min=0)  # 0.14 for primary only

    # convert p from physical to pixel coordinates
    size = np.array((128, 128, 128)).astype(int)
    spacing = np.array((2, 2, 2))
    offset = -size * spacing / 2.0 + spacing / 2.0
    print(f"Image size, spacing, offset: {size} {spacing} {offset}")
    pix = np.rint((positions - offset) / spacing).astype(int)

    # remove values out of the image fov
    print(f"Number of events:                   {m}")
    print(f"Number of events after E selection: {len(pix)}")
    for i in [0, 1, 2]:
        pix = pix[(pix[:, i] < size[i]) & (pix[:, i] > i)]
    print(f"Number of events in the image FOV:  {len(pix)}")

    # output filename
    if output == "AUTO":
        full_path = os.path.split(file_input)
        b, extension = os.path.splitext(full_path[1])
        if not output_folder:
            output_folder = "."
        output = f"{b}.mhd"
        output = os.path.join(output_folder, output)
    print(f"Output file: {output}")

    # create the image
    a = np.zeros(size)
    for x in pix:
        a[x[0], x[1], x[2]] += 1
    img = itk.image_from_array(a)
    itk.imwrite(img, output)


# --------------------------------------------------------------------------
if __name__ == "__main__":
    go()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import opengate as gate
from scipy.spatial.transform import Rotation

paths = gate.get_default_test_paths(__file__)

"""
Data :
gt_image_crop source_three_areas.mhd -o source_three_areas_crop.mhd
gt_affine_transform -i source_three_areas_crop.mhd -o source_three_areas_crop_3.5mm.mhd  --newspacing 3.5 -fr -a -im NN
"""

# create the simulation
sim = gate.Simulation()

# main options
ui = sim.user_info
ui.g4_verbose = False
ui.g4_verbose_level = 1
ui.visu = False
ui.number_of_threads = 1
ui.random_seed = 123456789
activity_bq = 1e6

# visu
if ui.visu:
    ui.number_of_threads = 1
    activity_bq = 100

# add a material database
sim.add_material_database(paths.data / "GateMaterials.db")

# units
m = gate.g4_units("m")
cm = gate.g4_units("cm")
mm = gate.g4_units("mm")
keV = gate.g4_units("keV")
MeV = gate.g4_units("MeV")
Bq = gate.g4_units("Bq")
gcm3 = gate.g4_units("g/cm3")

#  change world size
world = sim.world
world.size = [2 * m, 2 * m, 2 * m]

# add a fake object to test position & rotation
fake = sim.add_volume("Box", "fake")
fake.size = [1 * m, 1 * m, 1 * m]
fake.translation = [5 * cm, 10 * cm, 0 * cm]
r = Rotation.from_euler("y", 33, degrees=True)
fake.rotation = r.as_matrix()
fake.material = "G4_AIR"
fake.color = [0, 0, 1, 1]  # blue

# add an image
f = paths.data / "ct_4mm.mhd"
if ui.visu:
    f = paths.data / "ct_40mm.mhd"

ct = sim.add_volume("Image", "ct")
ct.image = f
ct.mother = "fake"
ct.material = "G4_AIR"
tol = 0.1 * gcm3
ct.voxel_materials, materials = gate.HounsfieldUnit_to_material(
    tol,
    paths.data / "Schneider2000MaterialsTable.txt",
    paths.data / "Schneider2000DensitiesTable.txt",
)
ct.dump_label_image = paths.output / "ct_4mm_labels.mhd"

# condGAN source with voxelized condition
source = sim.add_source("GANSource", "source")
source.mother = "ct"
source.cond_image = paths.data / "source_three_areas_crop_3.5mm.mhd"
source.position.translation = gate.get_translation_between_images_center(
    str(ct.image), str(source.cond_image)
)
source.particle = "alpha"
source.activity = activity_bq * Bq / ui.number_of_threads
source.compute_directions = True
source.pth_filename = paths.data / "train_gaga_v001_GP_0GP_10_60000.pth"
source.position_keys = ["PrePosition_X", "PrePosition_Y", "PrePosition_Z"]
# source.backward_distance = 20 * cm
# source.backward_force = True
source.cond_debug = True
source.direction_keys = ["PreDirection_X", "PreDirection_Y", "PreDirection_Z"]
source.energy.mono = 1 * MeV
source.energy_key = None
source.weight_key = None
source.time_key = None
source.relative_timing = True
source.batch_size = 1e5
source.verbose_generator = True

# cuts (not need precision here)
c = sim.set_cut("world", "all", 100 * mm)

# add dose actor
dose = sim.add_actor("DoseActor", "dose")
dose.output = paths.output / "test047-edep.mhd"
dose.mother = "ct"
dose.size = [70, 70, 240]
dose.spacing = [3.5 * mm, 3.5 * mm, 3.5 * mm]
dose.img_coord_system = True
dose.hit_type = "random"

# add stat actor
stats = sim.add_actor("SimulationStatisticsActor", "Stats")
stats.track_types_flag = True

# start simulation
# output = sim.start(True)
# FIXME
output = sim.start()

# ---------------------------------------------------------------
# print results at the end
print()
gate.warning("Tests stats file")
stat = output.get_actor("Stats")
print(stat)
ref_stat_file = paths.output_ref / "t047_stats.txt"
# stat.write(ref_stat_file) # (for reference)
stats_ref = gate.read_stat_file(ref_stat_file)
is_ok = gate.assert_stats(stat, stats_ref, 0.005)

print()
gate.warning("Compare image to analog")
is_ok = (
    gate.assert_images(
        paths.output_ref / "test047-edep.mhd",
        dose.output,
        stat,
        tolerance=19,
        ignore_value=0,
        axis="x",
    )
    and is_ok
)

print("Test with vv: ")
print(f"vv {source.cond_image} --fusion {dose.output}")

gate.test_ok(is_ok)



import opengate as gate
import opengate_core as g4


class TrdVolume(gate.VolumeBase):
    """
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html?highlight=g4trd

    dx1 Half-length along X at the surface positioned at -dz
    dx2 Half-length along X at the surface positioned at +dz
    dy1 Half-length along Y at the surface positioned at -dz
    dy2 Half-length along Y at the surface positioned at +dz
    zdz Half-length along Z axis

    """

    type_name = "Trd"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        u = user_info
        mm = gate.g4_units("mm")
        u.dx1 = 30 * mm
        u.dx2 = 10 * mm
        u.dy1 = 40 * mm
        u.dy2 = 15 * mm
        u.dz = 60 * mm

    def build_solid(self):
        u = self.user_info
        return g4.G4Trd(u.name, u.dx1, u.dx2, u.dy1, u.dy2, u.dz)



import opengate as gate
import opengate_core as g4


class HexagonVolume(gate.VolumeBase):
    type_name = "Hexagon"

    """
    This is a special case of a polyhedra
    https://geant4-userdoc.web.cern.ch/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html
    """

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        cm = gate.g4_units("cm")
        user_info.height = 5 * cm
        user_info.radius = 0.15 * cm

    def build_solid(self):
        u = self.user_info
        deg = gate.g4_units("deg")
        phi_start = 0 * deg
        phi_total = 360 * deg
        num_side = 6
        num_zplanes = 2
        zplane = [-u.height / 2, u.height / 2]
        radius_inner = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        radius_outer = [u.radius] * num_side

        return g4.G4Polyhedra(
            u.name,
            phi_start,
            phi_total,
            num_side,
            num_zplanes,
            zplane,
            radius_inner,
            radius_outer,
        )



import opengate as gate
import opengate_core as g4
import numpy as np


sigma_to_fwhm = 2 * np.sqrt(2 * np.log(2))
fwhm_to_sigma = 1.0 / sigma_to_fwhm


class DigitizerSpatialBlurringActor(
    g4.GateDigitizerSpatialBlurringActor, gate.ActorBase
):
    """
    Digitizer module for blurring a (global) spatial position.
    """

    type_name = "DigitizerSpatialBlurringActor"

    @staticmethod
    def set_default_user_info(user_info):
        gate.ActorBase.set_default_user_info(user_info)
        user_info.attributes = []
        user_info.output = "singles.root"
        user_info.input_digi_collection = "Hits"
        user_info.skip_attributes = []
        user_info.clear_every = 1e5
        user_info.blur_attribute = None
        user_info.blur_fwhm = None
        user_info.blur_sigma = None
        user_info.keep_in_solid_limits = True

    def __init__(self, user_info):
        # check and adjust parameters
        self.set_param(user_info)
        # base classes
        gate.ActorBase.__init__(self, user_info)
        if not hasattr(user_info.blur_sigma, "__len__"):
            user_info.blur_sigma = [user_info.blur_sigma] * 3
        g4.GateDigitizerSpatialBlurringActor.__init__(self, user_info.__dict__)
        actions = {"StartSimulationAction", "EndSimulationAction"}
        self.AddActions(actions)

    def set_param(self, user_info):
        if user_info.blur_fwhm is not None and user_info.blur_sigma is not None:
            gate.fatal(
                f"Error, use blur_sigma or blur_fwhm, not both "
                f"(there are: {user_info.blur_sigma} and {user_info.blur_fwhm}"
            )
        if user_info.blur_fwhm is not None:
            user_info.blur_sigma = np.array(user_info.blur_fwhm) * fwhm_to_sigma
        if user_info.blur_sigma is None:
            gate.fatal(f"Error, use blur_sigma or blur_fwhm")

    def __del__(self):
        pass

    def __str__(self):
        s = f"DigitizerSpatialBlurringActor {self.user_info.name}"
        return s

    def StartSimulationAction(self):
        g4.GateDigitizerSpatialBlurringActor.StartSimulationAction(self)

    def EndSimulationAction(self):
        g4.GateDigitizerSpatialBlurringActor.EndSimulationAction(self)



import opengate as gate
import opengate_core as g4


class TubsVolume(gate.VolumeBase):
    """
    http://geant4-userdoc.web.cern.ch/geant4-userdoc/UsersGuides/ForApplicationDeveloper/html/Detector/Geometry/geomSolids.html
    prmin SetInnerRadius
    prmax SetOuterRadius
    pdz  SetZHalfLength
    psphi SetStartPhiAngle
    pdphi SetDeltaPhiAngle
    """

    type_name = "Tubs"

    @staticmethod
    def set_default_user_info(user_info):
        gate.VolumeBase.set_default_user_info(user_info)
        # default values
        u = user_info
        mm = gate.g4_units("mm")
        deg = gate.g4_units("deg")
        u.rmin = 30 * mm
        u.rmax = 40 * mm
        u.dz = (
            40 * mm
        )  # Half-length in Z, while in previous Gate, the cylinder was total length (not half)
        u.sphi = 0 * deg
        u.dphi = 360 * deg

    def build_solid(self):
        u = self.user_info
        return g4.G4Tubs(u.name, u.rmin, u.rmax, u.dz, u.sphi, u.dphi)
